package net.sf.mgp.unclasses;

/*
 * Copyright 2008 MGP
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.FieldVisitor;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.signature.SignatureReader;
import org.objectweb.asm.signature.SignatureVisitor;

/**
 * Implementation of {@link DependencyResolver} resolving class generics type dependencies. Implemented using ObjectWeb ASM (http://asm.objectweb.org/).
 * 
 * <p>
 * This class should be removed as soon as released version of Apache BCEL will resolve generics types.
 * </p>
 * 
 * @plexus.component role="net.sf.mgp.unclasses.DependencyResolver" role-hint="generics"
 * 
 * @author Maksim Gordienko
 * 
 */
public class AsmClassGenericsTypeResolver extends AbstractAsmDependencyResolver {

	protected Collection getDependencyNames(ClassReader reader) {
		List types = new ArrayList();
		final SignatureVisitor signatureVisitor = new CollectingSignatureVisitor(types);

		reader.accept(new ClassAdapter(createEmptyClassVisitor()) {
			public void visit(int version, int access, String name, String signature, String superName, String[] interfaces) {
				visitSignature(signature, signatureVisitor);
			}

			public FieldVisitor visitField(int access, String name, String desc, String signature, Object value) {
				visitSignature(signature, signatureVisitor);
				return null;
			}

			public MethodVisitor visitMethod(int access, String name, String desc, String signature, String[] exceptions) {
				visitSignature(signature, signatureVisitor);
				return null;
			}
		}, 0);

		return types;
	}

	protected static void visitSignature(String signature, SignatureVisitor signatureVisitor) {
		if (signature != null) {
			SignatureReader r = new SignatureReader(signature);
			r.accept(signatureVisitor);
		}
	}

	private static class CollectingSignatureVisitor implements SignatureVisitor {

		private final Collection types;

		public CollectingSignatureVisitor(Collection types) {
			this.types = types;
		}

		public SignatureVisitor visitArrayType() {
			return this;
		}

		public void visitBaseType(char descriptor) {
		}

		public SignatureVisitor visitClassBound() {
			return this;
		}

		public void visitClassType(String name) {
			types.add(name.replace('/', '.'));
		}

		public void visitEnd() {
		}

		public SignatureVisitor visitExceptionType() {
			return this;
		}

		public void visitFormalTypeParameter(String name) {
		}

		public void visitInnerClassType(String name) {
		}

		public SignatureVisitor visitInterface() {
			return this;
		}

		public SignatureVisitor visitInterfaceBound() {
			return this;
		}

		public SignatureVisitor visitParameterType() {
			return this;
		}

		public SignatureVisitor visitReturnType() {
			return this;
		}

		public SignatureVisitor visitSuperclass() {
			return this;
		}

		public void visitTypeArgument() {
		}

		public SignatureVisitor visitTypeArgument(char wildcard) {
			return this;
		}

		public void visitTypeVariable(String name) {
		}
	}
}
