package net.sf.mgp.unclasses;

/*
 * Copyright 2008 MGP
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
import java.io.File;
import java.io.FileFilter;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Map.Entry;
import java.util.regex.Pattern;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.doxia.module.xhtml.decoration.render.RenderingContext;
import org.apache.maven.doxia.sink.Sink;
import org.apache.maven.doxia.site.decoration.Body;
import org.apache.maven.doxia.site.decoration.DecorationModel;
import org.apache.maven.doxia.siterenderer.Renderer;
import org.apache.maven.doxia.siterenderer.RendererException;
import org.apache.maven.doxia.siterenderer.SiteRenderingContext;
import org.apache.maven.doxia.siterenderer.sink.SiteRendererSink;
import org.apache.maven.doxia.tools.SiteTool;
import org.apache.maven.doxia.tools.SiteToolException;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;
import org.apache.maven.reporting.AbstractMavenReport;
import org.apache.maven.reporting.AbstractMavenReportRenderer;
import org.apache.maven.reporting.MavenReportException;
import org.codehaus.plexus.i18n.I18N;

/**
 * Unused classes goal
 * 
 * @goal unclasses
 * @execute phase="compile"
 * 
 * @author Maxim Gordienko
 * 
 */
public class UnclassesMojo extends AbstractMavenReport {
	private static final String FILE_SUFFIX = ".class";

	/**
	 * The input directory for the project classes.
	 * 
	 * @parameter expression="${project.build.outputDirectory}"
	 * @required
	 * @readonly
	 */
	private File classesDirectory;

	/**
	 * The output directory for the report.
	 * 
	 * @parameter expression="${project.reporting.outputDirectory}"
	 * @required
	 */
	private File outputDirectory;

	/**
	 * Do not report unused classes matching this pattern (regular expression). Pattern must match entire class name. By default report all classes.
	 * 
	 * @parameter default-value="^$"
	 */
	private String ignoreClasses;

	/**
	 * Internationalization component.
	 * 
	 * @component
	 */
	private I18N i18n;
	/**
	 * SiteTool component.
	 * 
	 * @since 2.1
	 * @component
	 */
	private SiteTool siteTool;
	/**
	 * Doxia Site Renderer component.
	 * 
	 * @component
	 */
	private Renderer siteRenderer;
	/**
	 * The Maven Project.
	 * 
	 * @parameter expression="${project}"
	 * @required
	 * @readonly
	 */
	private MavenProject project;

	/**
	 * Local Repository.
	 * 
	 * @parameter expression="${localRepository}"
	 * @required
	 * @readonly
	 */
	private ArtifactRepository localRepository;

	/**
	 * @component role="net.sf.mgp.unclasses.DependencyResolver"
	 */
	private List dependencyResolvers;

	private Map getUnusedClasses() throws MavenReportException {
		Pattern exclude = Pattern.compile(ignoreClasses);
		try {
			List files = getFiles(classesDirectory.getCanonicalPath());
			Map nodes = new HashMap();
			List excludedNodes = new ArrayList();
			for (Iterator it = files.iterator(); it.hasNext();) {
				String fileName = (String) it.next();
				String className = Utils.getClassNameFromFileName(fileName);
				GraphNode node = getNode(nodes, className);
				node.setProjectNode();
				if (exclude.matcher(className).matches()) {
					excludedNodes.add(node);
				}
				for (Iterator it2 = dependencyResolvers.iterator(); it2.hasNext();) {
					DependencyResolver resolver = (DependencyResolver) it2.next();
					Collection dependencies = resolver.getDependencyNames(classesDirectory, fileName);
					for (Iterator it3 = dependencies.iterator(); it3.hasNext();) {
						String dependency = (String) it3.next();
						GraphNode dependencyNode = getNode(nodes, dependency);
						node.addEdge(dependencyNode);
					}
				}
			}

			Map result = getUnusedClasses(nodes, excludedNodes);

			if (getLog().isDebugEnabled()) {
				getLog().debug("Unused classes: " + result);
			}
			return result;
		} catch (IOException e) {
			throw new MavenReportException("Execution failed", e);
		}
	}

	Map getUnusedClasses(Map nodes, List excludedNodes) {
		for (Iterator it = excludedNodes.iterator(); it.hasNext();) {
			GraphNode node = (GraphNode) it.next();
			nodes.remove(node.getValue());
			node.remove();
		}
		Map result = new TreeMap();
		for (Iterator it = nodes.values().iterator(); it.hasNext();) {
			GraphNode node = (GraphNode) it.next();
			if (node.isUnreferenced()) {
				Set dependants = new TreeSet();
				result.put(node.getValue(), dependants);
				for (Iterator it2 = node.getExclusiveProjectDependants().iterator(); it2.hasNext();) {
					GraphNode dependant = (GraphNode) it2.next();
					dependants.add(dependant.getValue());
				}
			}
		}
		return result;
	}

	private GraphNode getNode(Map nodes, String className) {
		GraphNode node = (GraphNode) nodes.get(className);
		if (node == null) {
			node = new GraphNode(className);
			nodes.put(className, node);
		}
		return node;
	}

	public void execute() throws MojoExecutionException {
		try {
			DecorationModel model = new DecorationModel();
			model.setBody(new Body());
			Map attributes = new HashMap();
			attributes.put("outputEncoding", "UTF-8");
			attributes.put("project", project);
			Locale locale = Locale.getDefault();
			Artifact defaultSkin = siteTool.getDefaultSkinArtifact(localRepository, project.getRemoteArtifactRepositories());
			SiteRenderingContext siteContext = siteRenderer.createContextForSkin(defaultSkin.getFile(), attributes, model, getName(locale), locale);

			RenderingContext context = new RenderingContext(outputDirectory, getOutputName() + ".html");

			SiteRendererSink sink = new SiteRendererSink(context);
			generate(sink, locale);

			outputDirectory.mkdirs();

			Writer writer = new OutputStreamWriter(new FileOutputStream(new File(outputDirectory, getOutputName() + ".html")), "UTF-8");

			siteRenderer.generateDocument(writer, sink, siteContext);
		} catch (IOException e) {
			throw new MojoExecutionException("Execution failed", e);
		} catch (SiteToolException e) {
			throw new MojoExecutionException("Execution failed", e);
		} catch (MavenReportException e) {
			throw new MojoExecutionException("Execution failed", e);
		} catch (RendererException e) {
			throw new MojoExecutionException("Execution failed", e);
		}
	}

	private List getFiles(String path) {
		List paths = new ArrayList();
		getFilesRecursively(path, "", paths);
		return paths;
	}

	private void getFilesRecursively(String path, String dirName, List paths) {
		File dir = new File(path);
		String[] files = dir.list(new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return name.endsWith(FILE_SUFFIX);
			}
		});
		for (int i = 0; i < files.length; i++) {
			paths.add(getFullPath(dirName, files[i]));
		}

		File[] childDirs = dir.listFiles(new FileFilter() {
			public boolean accept(File pathname) {
				return pathname.isDirectory();
			}
		});
		for (int i = 0; i < childDirs.length; i++) {
			getFilesRecursively(childDirs[i].getAbsolutePath(), getFullPath(dirName, childDirs[i].getName()), paths);
		}
	}

	private String getFullPath(String dir, String name) {
		if (dir.length() == 0) {
			return name;
		}
		return dir + File.separator + name;
	}

	/** {@inheritDoc} */
	protected String getOutputDirectory() {
		return outputDirectory.getAbsolutePath();
	}

	/** {@inheritDoc} */
	public File getReportOutputDirectory() {
		return outputDirectory;
	}

	/** {@inheritDoc} */
	protected MavenProject getProject() {
		return project;
	}

	/** {@inheritDoc} */
	protected Renderer getSiteRenderer() {
		return siteRenderer;
	}

	protected void executeReport(Locale locale) throws MavenReportException {
		ReportRenderer r = new ReportRenderer(getSink(), i18n, locale, getUnusedClasses());
		r.render();
	}

	public String getDescription(Locale locale) {
		return i18n.getString("unclasses-report", locale, "report.description");
	}

	public String getName(Locale locale) {
		return i18n.getString("unclasses-report", locale, "report.name");
	}

	public String getOutputName() {
		return "unclasses";
	}

	private static class ReportRenderer extends AbstractMavenReportRenderer {
		private final Locale locale;
		private final I18N i18n;
		private final Map classes;

		public ReportRenderer(Sink sink, I18N i18n, Locale locale, Map classes) {
			super(sink);
			this.i18n = i18n;
			this.locale = locale;
			this.classes = classes;
		}

		public String getTitle() {
			return i18n.getString("unclasses-report", locale, "report.title");
		}

		protected void renderBody() {
			startSection(i18n.getString("unclasses-report", locale, "report.unclasses.title"));
			paragraph(i18n.getString("unclasses-report", locale, "report.unclasses.description"));

			startTable();
			tableHeader(new String[] { i18n.getString("unclasses-report", locale, "report.class.name"),
					i18n.getString("unclasses-report", locale, "report.dependent.class.names") });
			for (Iterator it = classes.entrySet().iterator(); it.hasNext();) {
				Entry e = (Entry) it.next();
				sink.tableRow();
				tableCell((String) e.getKey());

				sink.tableCell();
				for (Iterator it2 = ((Collection) (e.getValue())).iterator(); it2.hasNext();) {
					String className = (String) it2.next();
					startTable();
					tableRow(new String[] { className });
					endTable();
				}
				sink.tableCell_();

				sink.tableRow_();
			}
			endTable();

			endSection();
		}

	}

}
